/** \file main.cpp
*  The main file.
*  Contains main function which calls functions to:
*  \li parse command line options
*  \li read in files and parse mesh
*  \li compute the circle pattern of the mesh
*  \li layout (and possibly cut) parameterized mesh
*/

#define _USE_MATH_DEFINES
#include <math.h>
#include <iostream>
#include <fstream>
#include <string>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "util.h"
#include "Mesh.h"
#include "Subdivide.h"

using namespace std;

void putOptionsInFromFile (int & argc, char * argv[], vector<string> & options);
void putOptionsIn         (int & argc, char * argv[], vector<string> & options);
void parseOptions (const vector<string> & options, ProgramOptions &po);
void checkOptions (ProgramOptions &po);

/**
The main function.
First function to be called
*/
int main (int argc, char *argv[]) {

   vector<string> options;

   if (argc < 2 || (argc != 2 && argc < 5))
      usage();

   if (argc == 2)
      putOptionsInFromFile(argc, argv, options);
   else
      putOptionsIn(argc, argv, options);

   clock_t startTime, finishTime;

   startTime = clock ();

   ProgramOptions po;
   parseOptions(options, po);
   checkOptions(po);

   for (int i = 0; i < po.num_steps; i++) {
      // Read in the mesh
      cout << "Parsing the mesh ";
      Mesh mesh;
      if (po.read_obj) {
	 cout << po.file_in << " (OBJ format)..." << endl;
	 mesh.readOBJ(po.file_in.c_str());
	 mesh.computeMeshInfo();
      }
      cout << "..Subdivision step " << (i+1) << "........." << endl;
      Subdivide subdivision(&mesh);

      if (po.method == 1)
	 subdivision.doMask1();
      else if (po.method == 2)
	 subdivision.doMask2();
      else if (po.method == 3)
	 subdivision.doMask3();
      else if (po.method == 4)
	 subdivision.doMask4();
      else {
	 cout << "Unknown method: " << po.method << endl;
	 cout << "Will perform first one." << endl;
	 subdivision.doMask1();
      }
      cout << "..........." << endl;
      if (po.write_obj) {
	 cout << "Writing out the results " << endl;
	 cout << po.obj_file_out << " (OBJ format)..." << endl;
	 subdivision.outputSubdSurface(po.obj_file_out.c_str());
      }

      po.file_in = po.obj_file_out;
   }

   finishTime = clock();
   cout << "\nFinnished, total: " 
      <<  ((double(finishTime)-double(startTime))/CLOCKS_PER_SEC) << " seconds." << endl;

}



void putOptionsInFromFile (int & argc, char * argv[], vector<string> & options) {
   ifstream in (argv[1]);
   check_error(!in, "Cannot open input file with options: " + string(argv[1]));
   string option;
   while (!in.eof()) {
      in >> ws;
      if (in.peek() == '\"') {
         in.ignore(1);
	 getline(in, option, '\"');
	 in.ignore(1);
      }
      else
	 in >> option;
      options.push_back(option);
   }
}

void putOptionsIn (int & argc, char * argv[], vector<string> & options) {
   for (int i = 1; i < argc; i++) {
      options.push_back(string(argv[i]));
   }
}

void checkOptions (ProgramOptions &po) {
   check_error((!po.read_obj), "No input file specified." );
   check_error((!po.write_obj), "No output file specified." ); 
   check_error((po.num_steps < 1), "Cannot do zero or negative number of subdivision steps." ); 
   check_warn((po.num_steps > 5), "Many subdivision steps specified, can take long time and/or memory." ); 

}

void parseOptions (const vector<string> & options, ProgramOptions &po) {
   for (unsigned int i = 0; i < options.size(); i++) {
      if (!strcmp(options[i].c_str(),"-io")){
	 po.read_obj = true;
	 check_error(i >= (options.size() - 1), string("No value for option speficied: ")+options[i]);
	 po.file_in = options[++i].c_str();

      }
      else if (!strcmp(options[i].c_str(),"-oo")){
	 po.write_obj = true;
	 check_error(i >= (options.size() - 1), string("No value for option speficied: ")+options[i]);
	 po.obj_file_out = options[++i].c_str();
      }
      else if (!strcmp(options[i].c_str(),"-n")){
	 check_error(i >= (options.size() - 1), string("No value for option speficied: ")+options[i]);
	 char * pEnd;
	 po.num_steps = strtol(options[++i].c_str(), &pEnd, 10);
      }
      else if (!strcmp(options[i].c_str(),"-m")){
	 check_error(i >= (options.size() - 1), string("No value for option speficied: ")+options[i]);
	 char * pEnd;
	 po.method = strtol(options[++i].c_str(), &pEnd, 10);
      }
      else {
	 cerr << "Warning, Unrecognized option: " << options[i].c_str() << endl;
      }
   }
}

